// $Id$

//=============================================================================
/**
 *  @file    util/process_mngr.h
 *
 *  @author  Fukasawa Mitsuo
 *
 *
 *    Copyright (C) 2004 BEE Co.,Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
//=============================================================================

#ifndef JYUGEM_BEE_UTIL_PROCESS_MNGR_H
#define JYUGEM_BEE_UTIL_PROCESS_MNGR_H

#include "jyugem/bee/beelib.h"
#include "ace/Process_Manager.h"
#include "ace/Thread.h"
#include "ace/Reactor.h"
#include "ace/Task.h"
#include <string>
#include <vector>
using namespace std;

namespace jyugem { namespace bee { namespace util
{

//
// Child process information
//
class child_process
{
public:
  //
  // Event Handler
  //
  class exit_handler : public ACE_Event_Handler
  {
  public:
    exit_handler(child_process * proc): m_proc(proc) {}

    virtual ~exit_handler(void) {}

    virtual int handle_close(ACE_HANDLE, ACE_Reactor_Mask)
    {
      delete this;
      return 0;
    }

    virtual int handle_exit(ACE_Process * p)
    {
      TRACE_INFO((_TX("exit(%s) got %d: %d\n"),
          m_proc->exename(), (int)(p->getpid()), (int)(p->exit_code())));
      m_proc->alive(false);
      return 0;
    }

  private:
    child_process * m_proc;
  };

  //
  // Constructor/Destructor
  //
  child_process(const std::string& pname)
    : m_alive(false), m_exename(pname), m_command(pname) {}
  child_process(const std::string& pname, const std::string& param)
    : m_alive(false), m_exename(pname)
  {
    m_command = pname;
    m_command += " ";
    m_command += param;
  }
  ~child_process() {}

  const BCHAR * exename() { return m_exename.c_str(); }
  const BCHAR * command() { return m_command.c_str(); }
  bool  alive()          { return m_alive; }
  void  alive(bool tf)   { m_alive = tf; }
  pid_t procid()         { return m_procid; }
  void  procid(pid_t id) { m_procid = id; }

private:
  child_process() : m_alive(false), m_exename(""), m_command("") {}

  // pid_t spawn(ACE_Process_Manager * mngr)
  // {
  //   ACE_Process_Options options;
  //   options.command_line(m_command.c_str());
  //   m_procid = mngr->spawn(this, options);
  //   if (m_procid != ACE_INVALID_PID)
  //   {
  //     TRACE_DEBUG((_TX("spawned child(%s): pid %d.\n"),
  //                  m_exename.c_str(), (int)m_procid));
  //   }
  //   else
  //   {
  //     TRACE_ERROR((_TX("%p : %s\n"), _TX("spawn failed"),
  //                  m_command.c_str()));
  //   }
  //   mngr->register_handler(new exit_handler(this), m_procid);
  //   m_alive = true;
  //   return m_procid;
  // }

protected:
  bool        m_alive;
  std::string m_exename;
  std::string m_command;
  pid_t       m_procid;
};


//
// Wrapper class of Process Manager
//
class process_mngr: public MT_Task
{
public:

  //
  // Default exit handler
  //
  class default_handler : public ACE_Event_Handler
  {
  public:
    default_handler(const BCHAR * msg = 0)
    {
      m_msg = (msg) ? msg : _TX("default");
    }
    virtual ~default_handler(void) {}

    virtual int handle_close(ACE_HANDLE, ACE_Reactor_Mask)
    {
      delete this;
      return 0;
    }

    virtual int handle_exit(ACE_Process * proc)
    {
      TRACE_INFO((_TX("exit(%s) got %d: %d\n"),
          m_msg.c_str(), (int)(proc->getpid()), (int)(proc->exit_code())));
      return 0;
    }

  private:
    string m_msg;
  };

  //
  // Constructor/Destructor
  //
  process_mngr(long interval_secs = 2, long delay_secs = 5)
      : m_abort(false), m_interval(interval_secs), m_delay(delay_secs)
  {
    m_mngr = ACE_Process_Manager::instance();
    m_mngr->open(ACE_Process_Manager::DEFAULT_SIZE, ACE_Reactor::instance());
    m_mngr->register_handler(new default_handler());
  }

  ~process_mngr(void) {}

  // Activate the thread.
  virtual int open(void * args = 0)
  {
    // Make this an Active Object.
    return this->activate();
  }

  virtual int close(u_long args = 0)
  {
    TRACE_INFO((_TX("close running\n")));
    for (size_t i = 0; i < m_procs.size(); i++)
    {
      if (m_procs[i] == 0)
        continue;

      if (m_procs[i]->alive() && m_procs[i]->procid() != ACE_INVALID_PID)
      {
        m_mngr->terminate(m_procs[i]->procid());
        m_procs[i]->alive(false);
      }
      delete m_procs[i];
      m_procs[i] = 0;

    }
    return 0;
  }

  int entry(const string& procname)
  {
    if (find(procname) != 0)
      return -1;

    child_process * proc = new child_process(procname);
    m_procs.push_back(proc);
    return 0;
  }

  int entry(const string& procname, const string& param)
  {
    if (find(procname) != 0)
      return -1;

    child_process * proc = new child_process(procname, param);
    m_procs.push_back(proc);
    return 0;
  }

  child_process * find(const string& procname)
  {
    for (size_t i = 0; i < m_procs.size(); i++)
    {
      if (procname == m_procs[i]->exename())
      {
        return m_procs[i];    // Find the process
      }
    }
    return 0;
  }

  int abort()
  {
    TRACE_FUNCTION();
    TRACE_DEBUG((_TX("abort processes\n")));
    m_abort = true;
    for (size_t i = 0; i < m_procs.size(); i++)
    {
      if (m_procs[i]->alive() && m_procs[i]->procid() != ACE_INVALID_PID)
      {
        TRACE_DEBUG((_TX("terminate child(%s): pid %d.\n"),
            m_procs[i]->exename(), (int)m_procs[i]->procid()));
        if (m_mngr->terminate(m_procs[i]->procid()) < 0)
        {
          TRACE_ERROR((_TX("%p : %d\n"), _TX("terminate failed"),
                       m_procs[i]->procid()));
        }
        m_procs[i]->alive(false);
      }
    }

    // Delay for stopping service.
    ACE_OS::sleep(m_delay + m_interval + 1);

    return 0;
  }

  int abort(const string& procname)
  {
    TRACE_FUNCTION();
    child_process * theProc = find(procname);
    if (theProc == 0)
      return -1;

    if (theProc->alive() && theProc->procid() != ACE_INVALID_PID)
    {
      TRACE_DEBUG((_TX("terminate child(%s): pid %d.\n"),
          theProc->exename(), (int)theProc->procid()));
      if (m_mngr->terminate(theProc->procid()) < 0)
      {
        TRACE_ERROR((_TX(" %p : %d\n"), _TX("terminate failed"),
                     theProc->procid()));
      }
      theProc->alive(false);
    }
    return 0;
  }

  pid_t restart(const string& procname)
  {
    child_process * theProc = find(procname);
    if (theProc == 0)
      return -1;

    spawn(theProc);
    return 0;
  }

  size_t prcount() { return m_mngr->managed(); }

//
private:
  pid_t spawn(const string& procname, const string& param)
  {
     child_process * proc = new child_process(procname, param);
     m_procs.push_back(proc);
     return this->spawn(proc);
  }

  pid_t spawn(child_process * child)
  {
    TRACE_FUNCTION();
    ACE_Process_Options options;
    options.command_line(child->command());
    pid_t pid = m_mngr->spawn(options);
    if (pid != ACE_INVALID_PID)
    {
      TRACE_DEBUG((_TX("spawned child(%s): pid %d.\n"), child->exename(),
                  (int)pid));

      child->procid(pid);
      //
      m_mngr->register_handler(new child_process::exit_handler(child), pid);
      //
      child->alive(true);
    }
    else
    {
      TRACE_ERROR((_TX("%p : %s\n"), ACE_TEXT("spawn failed"),
                   child->command()));
      child->procid(pid);
      child->alive(false);
    }
    return pid;
  }

  //
  // Service thread
  //
  virtual int svc(void)
  {
    TRACE_FUNCTION();
    log_manager::start_thread();

    size_t i;
    pid_t  pid;
    ACE_Time_Value tv;
    for (i = 0; i < m_procs.size(); i++)
    {
      pid = this->spawn(m_procs[i]);
      if (pid == ACE_INVALID_PID)
      {
        // this->close();
        return 0;                        // abort service
      }
    }

    ACE_OS::sleep(1);

    // Supervise to start processes
    tv = m_delay;
    pid = m_mngr->wait(0, tv);
    if (pid != 0)
    { // Error when process start
      TRACE_ERROR((_TX("Abort child(%d).\n"), (int)pid));
      // this->close();
      return 0;                         // abort service
    }

    // supervise alive processes
    tv = m_interval;
    while (1)
    {
      pid_t retid = m_mngr->wait(0, tv);
      if (retid == 0)
      {
        continue;
      }
      if (m_abort)
      {
        ACE_OS::sleep(m_delay);
        break;
      }

      // TRACE_DEBUG((_TX("killed child(%d).\n"), (int)retid));
      for (i = 0; i < m_procs.size(); i++)
      {
        if (m_procs[i]->procid() == retid)
        {
          pid = this->spawn(m_procs[i]);
          if (pid == ACE_INVALID_PID)
          {
            TRACE_ERROR((_TX("%p : %s\n"), _TX("re-spawn failed"),
                        m_procs[i]->command()));
          }
        }
      }
    }

    // this->close();
    return 0;
  }

private:
  bool m_abort;
  long m_interval;
  long m_delay;
  ACE_Process_Manager *   m_mngr;
  vector<child_process *> m_procs;

};


} } } // *namespace* jyugem::bee::util

#endif  // JYUGEM_BEE_UTIL_PROCESS_MNGR_H
