// $Id$

//////////////////////////////////////////////////////////////////////////////
// Copyright (c) Andreas Huber Doenni 2002-2004.
// Use, modification and distribution are subject to the Boost Software
// License, Version 1.0. (See accompanying file LICENSE_1_0.txt or copy at
// http://www.boost.org/LICENSE_1_0.txt)
//////////////////////////////////////////////////////////////////////////////

//=============================================================================
/**
 *  @file    ControlJobState.h
 *
 *  @author  Fukasawa Mitsuo
 *
 *
 *    Copyright (C) 2004 BEE Co.,Ltd. All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 */
//=============================================================================

//////////////////////////////////////////////////////////////////////////////
// The following code implements the state-machine:
//
// E94 : PROVISIONAL SPECIFICATION FOR CONTROL JOB MANAGEMENT
//
// Control Job State Model
//
// +---------------------------------------------------+
// |  O---+                              CONTROLJOB    |
// |      V *1                                         |
// |   +----------+        *2                          |
// |   |  QUEUED  |---------------------------------------->O
// |   |          |--+                                 |
// |   +----------+  |*3                               |
// |    *4  ^        |                                 |
// |  +-----|--------|-----------------------------+   |
// |  |     |        V                   ACTIVE    |   |
// |  |  +----------------+                        |   |
// |  |  |    SELECTED    |                        |   |
// |  |  |                |                        |   |
// |  |  +----------------+                        |   |
// |  |                |                           |   |
// |  | +---------+    |                           |   |
// |  | | WAITING | *6 |                           |   |
// |  | |   FOR   |<--(C)                          |   |
// |  | |  START  |    |                           |   |
// |  | +---------+    |*5                         |   |
// |  |    *7 |        |                           |   |
// |  |       V        V                           |   |
// |  |  +---------------+  *8   +--------------+  |   |
// |  |  |   EXECUTING   |------>|    PAUSE     |  |   |
// |  |  |               |<------|              |  |   |
// |  |  +---------------+  *9   +--------------+  |   |
// |  |               |                            |   |
// |  +---------------|----------------------------+   |
// |               *10|  *11|     | *12                |
// |                  V     V     V                    |
// |               +-----------------+     *13         |
// |               |    COMPLETED    |--------------------->O
// |               |                 |                 |
// |               +-----------------+                 |
// +---------------------------------------------------+
//
//  *1: Receive "Create" command.
//  *2: Receive "Cancel", "Abort" or "Stop" command.
//  *3: The processing resource has capacity to begin work on the next ControlJob.
//  *4: Receive "De-select" command.
//  *5: Material for the first process job arrives or in the case where the first
//      (or only) process job does not require material, this transition shall be
//      taken as soon as the processing resource for that process job becomes
//      available. "StartMethod" attribute in the ControlJob is set for AUTO.
//  *6: Same ad for transitions 5 except that the "StartMethod" attribute in the
//      ControlJob is set for USER_START.
//  *7: User START command received.
//  *8: Receive "Pause" command
//  *9: Receive "Resume" command
// *10: All the ProcessJobs specified for the ControlJob have completed.
// *11: Receive "CJStop" command.
// *12: Receive "CJAbort" command.
// *13: The ControlJob is deleted.
//

#ifndef JYUGEM_CJM_CONTROLJOBSTATE_H
#define JYUGEM_CJM_CONTROLJOBSTATE_H

#ifdef _MSC_VER
#pragma warning( disable: 4267 )
#endif

#include "jyugem/bee/state_machine_task.h"
#include "jyugem/bee/viewed_state.h"
namespace bee = jyugem::bee;

struct EvAbort    : fsm::event< EvAbort > {};
struct EvCancel   : fsm::event< EvCancel > {};
struct EvDeselect : fsm::event< EvDeselect > {};
struct EvHOQ      : fsm::event< EvHOQ > {};         // Head Of Queue
struct EvPause    : fsm::event< EvPause > {};
struct EvResume   : fsm::event< EvResume > {};
struct EvStart    : fsm::event< EvStart > {};
struct EvStop     : fsm::event< EvStop  > {};
struct EvBegin    : fsm::event< EvBegin > {};
struct EvComplete : fsm::event< EvComplete > {};


//
namespace jyugem { namespace cjm
{

typedef bee::named_state    NamedState;
typedef bee::viewed_state<> ViewedState;

//
// State Machine
//
class ControlJob;
class Queued;
class ControlJobState : public bee::state_machine_task< ControlJobState, Queued >
{
public:
  typedef bee::state_machine_task< ControlJobState, Queued >  task_base;

  ControlJobState(ControlJob * cjm);
  ~ControlJobState() {}

  virtual int send(trigger * trig)
  {
    return send_trigger<ViewedState>(trig->name());
  }

  // const std::string& current_name() { return this->current_state_name<ViewedState>(); }

  static void InitEventDictionary();
  static task_base::event_dictionary _event_dictionary;

  bool m_StartMode;
};


//
// Innor State
//
class Selected;
class Completed;
class Queued;
class InitActive;
class WaitingForStart;
class Executing;
class Pause;
class Completed;

//
//-----------------------------------------------------------------------------
//
class Completed : public fsm::state< Completed, ControlJobState >,
  public ViewedState
{
public:
  Completed(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new trigger("Complete");
    outermost_context().changed(trig);
  }
};

//-----------------------------------------------------------------------------
//
class Queued : public fsm::simple_state< Queued, ControlJobState, mpl::list<
  fsm::transition< EvBegin, Selected >, fsm::transition< EvAbort, Completed >,
  fsm::custom_reaction< EvStop > > >,
  public ViewedState
{
public:
  Queued() : ViewedState() { TRACE_STATE_ENTER(); }

  fsm::result react( const EvStop & )
  {
    // return transit< Completed >( &Completed::end, evt );
    return discard_event();
  }
};

//
//-----------------------------------------------------------------------------
//
class Active : public fsm::simple_state< Active, ControlJobState, mpl::list<
  fsm::transition< EvAbort, Completed>,
  fsm::transition< EvStop, Completed> >, InitActive >,
  public NamedState
{
public:
  Active() : NamedState() { TRACE_STATE_ENTER(); }
};

//
//-----------------------------------------------------------------------------
//
class InitActive : public fsm::state< InitActive, Active,
  fsm::custom_reaction< EvEvalCondition > >,
  public NamedState
{
public:
  InitActive(my_context ctx) : NamedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
  }

  fsm::result react( const EvEvalCondition & )
  {
    // return transit< Selected >();
    return discard_event();
  }
};

//
//-----------------------------------------------------------------------------
//
class Selected : public fsm::state< Selected, Active,
  fsm::custom_reaction< EvEvalCondition > >,
  public ViewedState
{
public:
  Selected(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    post_event(boost::intrusive_ptr< EvEvalCondition >(new EvEvalCondition()));
  }

  fsm::result react( const EvEvalCondition & );
};

//
//-----------------------------------------------------------------------------
//
class WaitingForStart : public fsm::state< WaitingForStart,
  Active, fsm::transition< EvStart, Executing > >,
  public ViewedState
{
public:
  WaitingForStart(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    bee::trigger * trig = new bee::trigger("Job Waiting For Start");
    outermost_context().changed(trig);
  }
};

//
//-----------------------------------------------------------------------------
//
class Executing : public fsm::state< Executing, Active, mpl::list<
  fsm::transition< EvComplete, Completed >,
  fsm::transition< EvPause, Pause > > >,
  public ViewedState
{
public:
  Executing(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();
    if (outermost_context().prevState(_TX("Paused")))
    {
      bee::trigger * trig = new bee::trigger("Resumed");
      outermost_context().changed(trig);
    }
    else
    {
      bee::trigger * trig = new bee::trigger("Executing began");
      outermost_context().changed(trig);
    }
  }
};

//
//-----------------------------------------------------------------------------
//
class Pause : public fsm::state< Pause, Active,
  fsm::transition< EvResume, Executing > >,
  public ViewedState
{
public:
  Pause(my_context ctx) : ViewedState(), my_base(ctx)
  {
    TRACE_STATE_ENTER();

    bee::trigger * trig = new bee::trigger("Paused");
    outermost_context().changed(trig);
  }
};


} }  // *namespace* jyugem::cjm

#endif  /* JYUGEM_CJM_CONTROLJOBSTATE_H */
